# -*- coding: utf-8 -*-
"""This module defines base class :class:`Atomic` that all other
:mod:`~prody.atomic` classes are derived from."""

from numpy import all, arange
from os import path
from prody import LOGGER, __path__
from prody.utilities import openData

from . import flags
from .bond import trimBonds
from .fields import READONLY

__all__ = ['Atomic', 'AAMAP']

NOTALLNONE = set(['not', 'all', 'none', 'index', 'sequence', 'x', 'y', 'z'])

MODMAP = {}
with openData('mod_res_map.dat') as f:
    for line in f:
        try:
            mod, aa = line.strip().split(' ')
            MODMAP[mod] = aa
        except:
            continue

CORE_AAMAP = AAMAP = {
    'ALA': 'A', 'ARG': 'R', 'ASN': 'N', 'ASP': 'D', 'CYS': 'C', 'GLN': 'Q',
    'GLU': 'E', 'GLY': 'G', 'HIS': 'H', 'ILE': 'I', 'LEU': 'L', 'LYS': 'K',
    'MET': 'M', 'PHE': 'F', 'PRO': 'P', 'SER': 'S', 'THR': 'T', 'TRP': 'W',
    'TYR': 'Y', 'VAL': 'V'
}

invAAMAP = dict((v, k) for k, v in CORE_AAMAP.items())

AAMAP = {
    'ALA': 'A', 'ARG': 'R', 'ASN': 'N', 'ASP': 'D', 'CYS': 'C', 'GLN': 'Q',
    'GLU': 'E', 'GLY': 'G', 'HIS': 'H', 'ILE': 'I', 'LEU': 'L', 'LYS': 'K',
    'MET': 'M', 'PHE': 'F', 'PRO': 'P', 'SER': 'S', 'THR': 'T', 'TRP': 'W',
    'TYR': 'Y', 'VAL': 'V',
    'ASX': 'B', 'GLX': 'Z', 'SEC': 'U', 'PYL': 'O', 'XLE': 'J', '': '-',
    'UNK': 'X'
}

# add bases
AAMAP.update({'ADE': 'a', 'THY': 't', 'CYT': 'c',
              'GUA': 'g', 'URA': 'u'})

# add reversed mapping
_ = {}
for aaa, a in AAMAP.items():
    _[a] = aaa
AAMAP.update(_)

# add modified AAs and bases to AAMAP
MODAAMAP = {}
for mod, aa in MODMAP.items():
    if aa in AAMAP:
        MODAAMAP[mod] = AAMAP[aa]
AAMAP.update(MODAAMAP)

class Atomic(object):

    """Base class for all atomic classes that can be used for type checking."""

    __slots__ = []

    def __getattribute__(self, name):

        try:
            return object.__getattribute__(self, name)

        except AttributeError:
            if name.startswith('is') and self.isFlagLabel(name[2:]):
                return all(self._getFlags(name[2:]))
            else:
                if name == 'all':
                    try:
                        ag = self.getAtomGroup()
                    except AttributeError:
                        ag = self
                        selstr = name
                        return Selection(ag, arange(self.numAtoms()), 'all',
                                         self._acsi, unique=True)
                    else:
                        try:
                            dummies = self.numDummies()
                        except AttributeError:
                            return Selection(ag, self.getIndices(),
                                             self.getSelstr(),
                                             self._acsi, unique=True)
                        else:
                            return AtomMap(ag, self.getIndices(), self._acsi,
                                           intarrays=True, dummies=dummies,
                                           title=self.getTitle())
                elif name == 'none':
                    return None
                elif self.isFlagLabel(name):
                    try:
                        ag = self.getAtomGroup()
                    except AttributeError:
                        ag = self
                        selstr = name
                    else:
                        selstr = '({0}) and ({1})'.format(name,
                                                          self.getSelstr())
                    try:
                        dummies = self.numDummies()
                    except AttributeError:
                        indices = self._getSubset(name)
                        if len(indices):
                            return Selection(ag, indices, selstr,
                                             self._acsi, unique=True)
                        else:
                            return None
                    else:
                        indices = self._getSubset(name)
                        if len(indices):
                            return AtomMap(ag, indices, self._acsi,
                                           intarrays=True, dummies=dummies,
                                           title='Selection ' + repr(name) +
                                                 ' from ' + str(self))
                        else:
                            return None
                elif name == '_anisous':
                    return None
                else:
                    selstr = name
                    items = name.split('_')
                    word = items[0]
                    if (self.isFlagLabel(word) or self.isDataLabel(word) or
                       word in NOTALLNONE or isSelectionMacro(word)):
                        selstr = ' '.join(items)
                        return SELECT.select(self, selstr)

        raise AttributeError('{0} object has no attribute `{1}` and {2} '
                             'is not a valid selection string'
                             .format(self.__class__.__name__, name,
                                     repr(selstr)))

    def __getstate__(self):

        return dict([(slot, getattr(self, slot))
                     for slot in self.__class__.__slots__])

    def __setstate__(self, state):

        for slot in self.__class__.__slots__:
            try:
                value = state[slot]
            except KeyError:
                pass
            else:
                setattr(self, slot, value)

    def copy(self):
        """Returns a copy of atoms (and atomic data) in an :class:`.AtomGroup`
        instance."""

        dummies = None
        indices = None
        readonly = False
        try:
            ag = self.getAtomGroup()
        except AttributeError:
            ag = self
            readonly = True
            new = AtomGroup(ag.getTitle())
        else:
            indices = self.getIndices()
            new = AtomGroup(ag.getTitle() + ' ' + str(self))
            try:
                dummies = self.numDummies()
            except AttributeError:
                pass
            else:
                if dummies:
                    dummy = self.getFlags('dummy')
                    mapped = self.getFlags('mapped')

        try:
            self.getIndex()
        except AttributeError:
            this = self
        else:
            this = self.all

        if self.numCoordsets():
            new.setCoords(this.getCoordsets(), label=ag.getCSLabels())

        for label in ag.getDataLabels():
            if label in READONLY:
                if readonly:
                    new._data[label] = this.getData(label)
            else:
                new.setData(label, this.getData(label))

        #if readonly:
        #    for label in READONLY:
        #        data = this.getData(label)
        #        if data is not None:
        #            new._data[label] = data

        skip_flags = set()
        for label in ag.getFlagLabels():
            if label in skip_flags:
                continue
            else:
                new._setFlags(label, this.getFlags(label))
                skip_flags.update(flags.ALIASES.get(label, [label]))

        if dummies:
            new._setFlags('dummy', dummy)
            new._setFlags('mapped', mapped)

        bonds = ag._bonds
        bmap = ag._bmap
        if bonds is not None and bmap is not None:
            if indices is None:
                new._bonds = bonds.copy()
                new._bmap = bmap.copy()
                new._data['numbonds'] = ag._data['numbonds'].copy()
            elif dummies:
                if dummies:
                    indices = indices[self._getMapping()]
                if len(set(indices)) == len(indices):
                    new.setBonds(trimBonds(bonds, indices))
                else:
                    LOGGER.warn('Duplicate atoms in mapping, bonds are '
                                'not copied.')
            else:
                bonds = trimBonds(bonds, indices)
                if bonds is not None:
                    new.setBonds(bonds)
        return new

    __copy__ = copy
    toAtomGroup = copy

    def select(self, selstr, **kwargs):
        """Returns atoms matching *selstr* criteria.  See :mod:`~.select` module
        documentation for details and usage examples."""

        return SELECT.select(self, selstr, **kwargs)

    def getTitle(self):
        """Returns title of the instance."""
        try:
            ag = self.getAtomGroup()
        except AttributeError:
            ag = self
        return ag._title
    
    def getSequence(self, **kwargs):
        """Returns one-letter sequence string for amino acids, unless *longSeq* is **True**.
        When *allres* keyword argument is **True**, sequence will include all
        residues (e.g. water molecules) in the chain and **X** will be used for
        non-standard residue names."""

        longSeq = kwargs.get('longSeq', False)

        get = AAMAP.get
        if hasattr(self, 'getResnames'):
            if longSeq:
                seq = ' '.join(self.getResnames())
            else:
                seq = ''.join([get(res, 'X') for res in self.getResnames()])
        else:
            res = self.getResname()
            if longSeq:
                seq = res
            else:
                seq = get(res, 'X')
        
        return seq

    def toTEMPyAtoms(self):
        """Returns a :class:`TEMPy.protein.prot_rep_biopy.Atom` or list of them as appropriate"""
        try:
            from TEMPy.protein.prot_rep_biopy import Atom as TEMPyAtom
        except ImportError:
            raise ImportError('TEMPy is needed for this functionality')

        if hasattr(self, 'getResnums'):
            return [atom.toTEMPyAtom() for atom in self if atom is not None]
        else:
            return [self.toTEMPyAtom()]

    def toTEMPyStructure(self):
        """Returns a :class:`.protein.prot_rep_biopy.Structure` object""" 
        try:
            from TEMPy.protein.prot_rep_biopy import BioPy_Structure
        except ImportError:
            raise ImportError('TEMPy is needed for this functionality')

        return BioPy_Structure(self.toTEMPyAtoms())

    def numResidues(self):
        """Returns number of residues."""

        return len(set(self._getResindices()))


    def toBioPythonStructure(self, header=None, **kwargs):
        """Returns a :class:`Bio.PDB.Structure` object

        :arg atoms: an object with atom and coordinate data
        :type atoms: :class:`.Atomic`

        :arg csets: coordinate set indices, default is all coordinate sets
        """ 
        try:
            from Bio.PDB.StructureBuilder import StructureBuilder
            from Bio.PDB.PDBExceptions import PDBConstructionException
        except ImportError:
            raise ImportError('Bio StructureBuilder could not be imported. '
                'Reinstall ProDy or install Biopython '
                'to solve the problem.')

        origACSI = self.getACSIndex()

        csets = kwargs.get('csets', None)
        if csets is None:
            csets = range(self.numCoordsets())

        structure_builder = StructureBuilder()
        structure_builder.init_structure(self.getTitle())
        if header is not None:
            structure_builder.set_header(header)

        result = structure_builder.get_structure()
        result.is_pqr = (self.getCharges() is not None 
                         and self.getRadii() is not None)
        
        for i in csets:
            self.setACSIndex(i)
            structure_builder.init_model(i, i+1)

            current_segid = None
            current_chain_id = None
            current_residue_id = None

            for global_line_counter, atom in enumerate(self):
                segid = atom.getSegname()
                if current_segid != segid:
                    current_segid = segid
                    structure_builder.init_seg(current_segid)

                chainid = atom.getChid()
                resname = atom.getResname()

                if atom.getFlag('hetatm'):
                    if atom.getFlag('water'):
                        hetero_flag = 'W'
                    else:
                        hetero_flag = 'H'
                else:
                    hetero_flag = ' '

                resseq = atom.getResnum()
                icode = atom.getIcode()
                if len(icode) == 0:
                    icode = ' '
                residue_id = (hetero_flag, resseq, icode)

                if current_chain_id != chainid:
                    current_chain_id = chainid
                    structure_builder.init_chain(current_chain_id)
                    
                    current_residue_id = residue_id
                    current_resname = resname
                    try:
                        structure_builder.init_residue(
                            resname, hetero_flag, resseq, icode
                        )
                    except PDBConstructionException as message:
                        result._handle_PDB_exception(message, global_line_counter)
                elif current_residue_id != residue_id or current_resname != resname:
                    current_residue_id = residue_id
                    current_resname = resname
                    try:
                        structure_builder.init_residue(
                            resname, hetero_flag, resseq, icode
                        )
                    except PDBConstructionException as message:
                        result._handle_PDB_exception(message, global_line_counter)

                name = atom.getName()
                coord = atom.getCoords()
                altloc = atom.getAltloc()
                fullname = atom.getName()
                serial_number = atom.getSerial()
                element = atom.getElement()

                if not result.is_pqr:
                    # init atom with pdb fields
                    try:
                        structure_builder.init_atom(
                            name,
                            coord,
                            atom.getBeta(),
                            atom.getOccupancy(),
                            altloc,
                            fullname,
                            serial_number,
                            element,
                        )
                    except PDBConstructionException as message:
                        result._handle_PDB_exception(message, global_line_counter)
                else:
                    try:
                        structure_builder.init_atom(
                            name,
                            coord,
                            atom.getCharge(),
                            atom.getRadius(),
                            altloc,
                            fullname,
                            serial_number,
                            element,
                            atom.getCharge(),
                            atom.getRadius(),
                            result.is_pqr,
                        )
                    except PDBConstructionException as message:
                        result._handle_PDB_exception(message, global_line_counter)

                if atom.getAnisou() is not None:
                    structure_builder.set_anisou(atom.getAnisou())

                if atom.getAnistd() is not None:
                    structure_builder.set_siguij(atom.getAnistd())

        self.setACSIndex(origACSI)

        return result


# -*- coding: utf-8 -*-
"""This module defines :class:`Acceptor` for dealing with acceptor information provided
by using :meth:`.AtomGroup.setAcceptors` method."""

from numbers import Integral
import numpy as np

__all__ = ['Acceptor']

class Acceptor(object):

    """A pointer class for acceptored atoms.  Following built-in functions are
    customized for this class:

    * :func:`len` returns acceptor length, i.e. :meth:`getLength`
    * :func:`iter` yields :class:`~.Atom` instances"""

    __slots__ = ['_ag', '_acsi', '_indices']

    def __init__(self, ag, indices, acsi=None):

        self._ag = ag
        self._indices = np.array(indices)
        if acsi is None:
            self._acsi = ag.getACSIndex()
        else:
            self._acsi = acsi

    def __repr__(self):

        one, two = self._indices
        names = self._ag._getNames()
        return '<Acceptor: {0}({1})--{2}({3}) from {4}>'.format(
                            names[one], one, names[two], two, str(self._ag))

    def __str__(self):

        one, two = self._indices
        names = self._ag._getNames()
        return '{0}({1})--{2}({3})'.format(
                                            names[one], one, names[two], two)

    def __eq__(self, other):

        return (isinstance(other, Acceptor) and other.getAtomGroup() is self._ag
                and (np.all(other.getIndices() == self._indices) or
                 np.all(other.getIndices() == list(reversed(self._indices)))))

    def __ne__(self, other):

        return not self.__eq__(other)

    def __len__(self):

        return self.getLength()

    def __iter__(self):

        for index in self._indices:
            yield self._ag[index]

    def getAtomGroup(self):
        """Returns atom group."""

        return self._ag

    def getAtoms(self):
        """Returns acceptored atoms."""

        return (self._ag[self._indices[0]], self._ag[self._indices[1]])

    def getIndices(self):
        """Returns indices of acceptored atoms."""

        return self._indices.copy()

    def getLength(self):
        """Returns acceptor length."""

        vector = self.getVector()
        return np.multiply(vector, vector, vector).sum() ** 0.5

    def getVector(self):
        """Returns acceptor vector that originates from the first atom."""

        one, two = self._indices
        acsi = self.getACSIndex()
        return self._ag._coords[acsi, two] - self._ag._coords[acsi, one]

    def getACSIndex(self):
        """Returns index of the coordinate set."""

        acsi = self._acsi
        if acsi >= self._ag._n_csets:
            raise ValueError('{0} has fewer coordsets than assumed by {1}'
                             .format(str(self._ag), str(self)))
        return acsi

    def setACSIndex(self, index):
        """Set the coordinate set at *index* active."""

        if self._ag._coords is None:
            raise AttributeError('coordinates are not set')

        if not isinstance(index, Integral):
            raise TypeError('index must be an integer')

        n_csets = self._ag._n_csets
        if n_csets <= index or n_csets < abs(index):
            raise IndexError('coordinate set index is out of range')

        if index < 0:
            index += n_csets

        self._acsi = index


def evalAcceptors(acceptors, n_atoms):
    """Returns an array mapping atoms to their common acceptor neighbors and an array
    that stores number of acceptors made by each atom.
    
    This excludes entries corresponding to missing hydrogens (with 0 in them that became -1)."""

    if acceptors.min() < 0:
        # remove entries corresponding to missing hydrogens (with 0 in them that became -1)
        acceptors = acceptors[np.argwhere(acceptors == -1)[-1][0]+1:]

    numacceptors = np.bincount(acceptors.reshape((acceptors.shape[0] * 2)))

    acmap = np.zeros((n_atoms, numacceptors.max()), int)
    index = np.zeros(n_atoms, int)
    
    acmap.fill(-1)
    
    for acceptor in acceptors:
        a, b = acceptor
        acmap[a, index[a]] = b
        acmap[b, index[b]] = a
        index[acceptor] += 1
    return acmap, numacceptors


def trimAcceptors(acceptors, indices):
    """Returns acceptors between atoms at given indices."""

    iset = set(indices)
    acceptors = [acceptor for acceptor in acceptors if acceptor[0] in iset and acceptor[1] in iset]
    if acceptors:
        newindices = np.zeros(indices.max()+1, int)
        newindices[indices] = np.arange(len(indices))
        return newindices[np.array(acceptors)]



from prody.tests import TestCase

from numpy import array, log, zeros, char
from numpy.testing import assert_equal, assert_array_equal, assert_array_almost_equal

from prody.tests.datafiles import *

from prody import LOGGER, refineMSA, parseMSA, calcMSAOccupancy, mergeMSA
from prody import uniqueSequences

LOGGER.verbosity = None

FASTA = parseMSA(pathDatafile('msa_Cys_knot.fasta'))
FASTA_ALPHA = char.isalpha(FASTA._msa)
NUMSEQ = FASTA.numSequences() * 1.

class TestRefinement(TestCase):


    def testLabel(self):
        label = 'FSHB_BOVIN'
        index = FASTA.getIndex(label)
        refined = refineMSA(FASTA, label=label)._getArray()

        expected = FASTA._getArray().take(FASTA_ALPHA[index].nonzero()[0], 1)

        assert_array_equal(refined, expected)


    def testRowocc(self):

        refined = refineMSA(FASTA, rowocc=0.9)._getArray()
        expected = FASTA._getArray()[FASTA_ALPHA.sum(1) / 112. >= 0.9,:]

        assert_array_equal(refined, expected)

    def testColocc(self):

        refined = refineMSA(FASTA, colocc=0.9)._getArray()
        expected = FASTA._getArray()[:, FASTA_ALPHA.sum(0) / NUMSEQ >= 0.9]

        assert_array_equal(refined, expected)

    def testRowCol(self):

        rowocc = 0.9
        colocc = 1.0
        refined = refineMSA(FASTA, rowocc=rowocc,
                            colocc=colocc)._getArray()
        rows = FASTA_ALPHA.sum(1) / 112. >= rowocc
        expected = FASTA._getArray()[rows]
        cols = char.isalpha(expected).sum(0,
                    dtype=float) / expected.shape[0] >= colocc

        expected = expected.take(cols.nonzero()[0], 1)
        assert_array_equal(refined, expected)

    def testSeqid(self):

        seqid = 0.50
        label = 'FSHB_BOVIN'
        unique = uniqueSequences(FASTA, seqid)
        refined = refineMSA(FASTA, seqid=seqid)
        assert_array_equal(refined._getArray(), FASTA._getArray()[unique])

    def testSeqidLabel(self):

        seqid = 0.50
        label = 'FSHB_BOVIN'
        labeled = refineMSA(FASTA, label=label)
        unique = uniqueSequences(labeled, seqid)
        unique[FASTA.getIndex(label)] = True
        refined = refineMSA(FASTA, label=label, seqid=seqid)
        assert_array_equal(refined._getArray(), labeled._getArray()[unique])


    def testAll(self):

        rowocc = 0.9
        colocc = 0.9
        seqid = 0.98
        label = 'FSHB_BOVIN'
        refined = refineMSA(FASTA, label=label, seqid=seqid,
                            rowocc=rowocc, colocc=colocc)

        index = FASTA.getIndex(label)
        which = FASTA_ALPHA[index].nonzero()[0]
        expected = FASTA._getArray().take(which, 1)

        expected = expected[uniqueSequences(expected, seqid)]

        expected = expected[calcMSAOccupancy(expected, 'row') >= rowocc]

        which = (calcMSAOccupancy(expected) >= colocc).nonzero()[0]
        expected = expected.take(which, 1)

        assert_array_equal(refined._getArray(), expected)

    def testAddition(self):
        numSeq = FASTA.numSequences()
        msa = FASTA + FASTA
        assert_array_equal(msa[:numSeq].getArray(), FASTA.getArray(), 'MSA addition failed')

    def testExtension(self):
        numSeq = FASTA.numSequences()
        msa = FASTA[:]
        msa.extend(FASTA)
        assert_equal(msa[numSeq:].getArray(), FASTA.getArray(), 'MSA extension failed')

        seqs = [str(seq) for seq in FASTA]
        msa = FASTA[:]
        msa.extend(FASTA)
        assert_equal(msa[numSeq:].getArray(), FASTA.getArray(), 'MSA extension failed')

class TestMerging(TestCase):


    def testMerge(self):

        merged = mergeMSA(FASTA, FASTA)
        length = FASTA.numResidues()
        self.assertEqual(merged[:, :length], merged[:, length:])



# -*- coding: utf-8 -*-
"""This module defines functions for accessing wwPDB servers."""

from os import getcwd
from os.path import join

from prody import LOGGER
from prody.utilities import openURL, sympath

from .localpdb import pathPDBFolder

__all__ = ['fetchPDBfromOPM']

def fetchPDBfromOPM(pdb, filename=None):
    url = 'https://opm-assets.storage.googleapis.com/pdb/%s.pdb' % pdb.lower()

    try:
        handle = openURL(url)
    except Exception as err:
        LOGGER.warn('{0} download failed ({1}).'.format(pdb, str(err)))
        filename = ''
    else:
        data = handle.read()
        if len(data):

            if filename is None:
                output_folder = pathPDBFolder()
                if output_folder is None:
                    output_folder = getcwd()

                filename = join(output_folder, pdb + '-opm.pdb')

            with open(filename, 'w+b') as pdbfile:
                pdbfile.write(data)

            LOGGER.debug('{0} downloaded ({1})'
                            .format(pdb, sympath(filename)))
        else:
            LOGGER.warn('{0} download failed, reason unknown.'
                        .format(pdb))
            filename = ''

    return filename



# -*- coding: utf-8 -*-
"""This module defines MSA analysis functions."""

from numbers import Integral

from numpy import all, zeros, dtype, array, char, cumsum, reshape
from numpy import vstack, isscalar

from prody import LOGGER, PY3K
from prody.utilities import toChararray, pystr, splitSeqLabel
from .sequence import Sequence

import sys

__all__ = ['MSA', 'refineMSA', 'mergeMSA', 'specMergeMSA',]

class MSA(object):

    """Store and manipulate multiple sequence alignments."""

    def __init__(self, msa, title='Unknown', labels=None, **kwargs):
        """*msa* must be a 2D Numpy character array. *labels* is a list of
        sequence labels (or titles).  *mapping* should map label or part of
        label to sequence index in *msa* array. If *mapping* is not given,
        one will be build from *labels*."""

        self._aligned = aligned = kwargs.get('aligned', True)
        msa = toChararray(msa, aligned)
        numseq = msa.shape[0]

        if labels and len(labels) != numseq:
            raise ValueError('len(labels) must be equal to number of '
                             'sequences')
        
        if labels is None:
            labels = [str(i+1) for i in range(numseq)]

        if PY3K:
            for i, label in enumerate(labels):
                if not isinstance(label, str):
                    labels[i] = label.decode()
                    
        self._labels = labels
        
        mapping = kwargs.get('mapping')
        self._map(mapping)
        self._msa = msa
        self._title = str(title) or 'Unknown'
        self._split = bool(kwargs.get('split', True))

    def _map(self, mapping=None):

        labels = self._labels
        if mapping is not None:
            try:
                mapping['isdict']
            except KeyError:
                pass
            except Exception:
                raise TypeError('mapping must be a dictionary')
            
            for key, value in mapping.items():
                values = [value] if isscalar(value) else value
                for i in values:
                    if not key in labels[i]:
                        labels[i] = key
        
        self._mapping = mapping = {}
        for index, label in enumerate(labels):
            label = splitSeqLabel(label)[0]
            try:
                value = mapping[label]
            except KeyError:
                mapping[label] = index
            else:
                try:
                    value.append(index)
                except AttributeError:
                    mapping[label] = [value, index]
        return mapping

    def __str__(self):

        return 'MSA ' + self._title

    def __repr__(self):

        if self._aligned:
            return ('<MSA: {0} ({1} sequences, {2} residues)>'
                    ).format(self._title, self.numSequences(),
                             self.numResidues())
        else:
            return ('<MSA: {0} ({1} sequences, not aligned)>'
                    ).format(self._title, self.numSequences())

    def __len__(self):

        return len(self._msa)

    def __add__(self, other):
        """Concatenate two MSAs."""

        A = self.getArray()
        B = other.getArray()
        try:
            AB = vstack((A, B))
        except:
            raise ValueError('failed to add {0} and {1} together: shapes do not match'
                             .format(repr(self), repr(other)))
        aligned = self._aligned or other._aligned

        labels = list(self._labels)
        labels.extend(other._labels)

        split = self._split or other._split

        msa = MSA(AB, title='%s + %s'%(self.getTitle(), other.getTitle()), 
                  aligned=aligned, labels=labels, split=split)
        return msa

    def __getitem__(self, index):

        if isinstance(index, Integral):
            return Sequence(self, index)

        if isinstance(index, str):
            try:
                rows = self._mapping[index]
            except KeyError:
                raise KeyError('label {0} is not mapped to a sequence'
                               .format(index))
            else:
                msa = self._msa[rows]
                if msa.ndim == 1:
                    return Sequence(self, rows)
                else:
                    if msa.base is not None:
                        msa = msa.copy()
                    labels = self._labels
                    return MSA(msa, title='{0}[{1}]'.format(self._title,
                               index), labels=[labels[i] for i in rows])
        elif isinstance(index, tuple):
            if len(index) == 1:
                return self[index[0]]
            elif len(index) == 2:
                rows, cols = index
            else:
                raise IndexError('invalid index: ' + str(index))
        else:
            rows, cols = index, None

        # handle list of labels
        if isinstance(rows, list):
            rows = self.getIndex(rows) or rows
        elif isinstance(rows, int):
            return Sequence(self._msa[rows, cols].tobytes(),
                            self._labels[rows])
        elif isinstance(rows, str):
            try:
                rows = self._mapping[rows]
            except KeyError:
                raise KeyError('label {0} is not mapped to a sequence'
                               .format(index))
            else:
                if isinstance(rows, int):
                    return Sequence(self._msa[rows, cols].tobytes(),
                                    self._labels[rows])

        if cols is None:
            msa = self._msa[rows]
        else:
            try:
                msa = self._msa[rows, cols]
            except TypeError:
                raise IndexError('invalid index: ' + str(index))

        try:
            lbls = list(array(self._labels)[rows])
        except TypeError:
            labels = self._labels
            lbls = [labels[i] for i in rows]
        finally:
            if not isinstance(lbls, list):
                lbls = [lbls]

        if msa.ndim == 0:
            msa = msa.reshape((1, 1))
        elif msa.ndim == 1:
            msa = msa.reshape((1, len(msa)))
        if msa.base is not None:
            msa = msa.copy()

        return MSA(msa=msa, title=self._title, labels=lbls,
                   aligned=self._aligned)

    def __iter__(self):

        for i in range(len(self._msa)):
            yield Sequence(self, i)

    def __contains__(self, key):

        try:
            return key in self._mapping
        except Exception:
            pass
        return False

    def __eq__(self, other):

        try:
            other = other._getArray()
        except AttributeError:
            return False

        try:
            return all(other == self._msa)
        except Exception:
            pass
        return False

    def _getSplit(self):

        return self._split

    def _setSplit(self, split):

        self._split = bool(split)

    split = property(_getSplit, _setSplit,
                     doc='Return split label when iterating or indexing.')

    def extend(self, other):
        """Adds *other* to this MSA."""

        A = self.getArray()
        if isinstance(other, MSA):
            B = other.getArray()
            otherlabels = other._labels
        elif isinstance(other, Sequence):
            B = other.getArray()
            B = reshape(B, (1, B.shape[0]))
            otherlabels = other.getLabel(full=True)
        else:
            try:
                B = toChararray(other, aligned=False)
                numA, numB = A.shape[0], B.shape[0]
                otherlabels = [str(i+1) for i in range(numA, numA+numB)]
            except:
                raise ValueError('failed to add {1} to {0}'
                             .format(repr(self), repr(other)))
        try:
            AB = vstack((A, B))
        except:
            raise ValueError('failed to add {1} to {0}: shapes do not match'
                             .format(repr(self), repr(other)))

        labels = list(self._labels)
        labels.extend(otherlabels)

        self._msa = AB
        self._labels = labels
        self._map()

    def isAligned(self):
        """Returns **True** if MSA is aligned."""

        return self._aligned

    def numSequences(self):
        """Returns number of sequences."""

        return self._msa.shape[0]

    def numResidues(self):
        """Returns number of residues (or columns in the MSA), if MSA is
        aligned."""

        if self._aligned:
            return self._msa.shape[1]

    def numIndexed(self):
        """Returns number of sequences that are indexed using the identifier
        part or all of their labels.  The return value should be equal to
        number of sequences."""

        count = len(self._mapping)
        if len(self._msa) == count:
            return count
        else:
            count = len(self._mapping)
            for val in self._mapping.values():
                try:
                    count += len(val) - 1
                except TypeError:
                    pass
            return count

    def getTitle(self):
        """Returns title of the instance."""

        return self._title

    def setTitle(self, title):
        """Set title of the instance."""

        self._title = str(title)

    def getLabel(self, index, full=False):
        """Returns label of the sequence at given *index*.  Residue numbers will
        be removed from the sequence label, unless *full* is **True**."""

        index = self._mapping.get(index, index)
        if full:
            return self._labels[index]
        else:
            return splitSeqLabel(self._labels[index])[0]

    def getResnums(self, index):
        """Returns starting and ending residue numbers (:term:`resnum`) for the
        sequence at given *index*."""

        index = self._mapping.get(index, index)
        return splitSeqLabel(self._labels[index])[1:]

    def getArray(self):
        """Returns a copy of the MSA character array."""

        return self._msa.copy()

    def _getArray(self):
        """Returns MSA character array."""

        return self._msa

    def getIndex(self, label):
        """Returns index of the sequence that *label* maps onto.  If *label*
        maps onto multiple sequences or *label* is a list of labels, a list
        of indices is returned.  If an index for a label is not found,
        return **None**."""

        try:
            index = self._mapping[label]
        except KeyError:
            try:
                return [v for k, v in self._mapping.items() if label in k][0]
            except:
                return None
        except TypeError:
            mapping = self._mapping
            indices = []
            append, extend = indices.append, indices.extend
            for key in label:
                try:
                    index = mapping[key]
                except KeyError:
                    return None
                try:
                    extend(index)
                except TypeError:
                    append(index)
            return indices
        else:
            try:
                return list(index)
            except TypeError:
                return index

    def iterLabels(self, full=False):
        """Yield sequence labels.  By default the part of the label used for
        indexing sequences is yielded."""

        if full:
            for label in self._labels:
                yield label
        else:
            for label in self._labels:
                yield splitSeqLabel(label)[0]

    def getLabels(self, full=False):
        """Returns all labels"""

        labels = [label for label in self.iterLabels(full=full)]
        return labels

    def countLabel(self, label):
        """Returns the number of sequences that *label* maps onto."""

        try:
            return len(self._mapping[label])
        except KeyError:
            return 0
        except TypeError:
            return 1

def refineMSA(msa, index=None, label=None, rowocc=None, seqid=None, colocc=None, **kwargs):
    """Refine *msa* by removing sequences (rows) and residues (columns) that
    contain gaps.

    :arg msa: multiple sequence alignment
    :type msa: :class:`.MSA`

    :arg index: remove columns that are gaps in the sequence with that index
    :type index: int

    :arg label: remove columns that are gaps in the sequence matching label,
        ``msa.getIndex(label)`` must return a sequence index, a PDB identifier
        is also acceptable
    :type label: str

    :arg rowocc: row occupancy, sequences with less occupancy will be
        removed after *label* refinement is applied
    :type rowocc: float

    :arg seqid: keep unique sequences at specified sequence identity level,
        unique sequences are identified using :func:`.uniqueSequences`
    :type seqid: float

    :arg colocc: column occupancy, residue positions with less occupancy
        will be removed after other refinements are applied
    :type colocc: float

    :arg keep: keep columns corresponding to residues not resolved in the PDB
        structure, default is **False**, applies when *label* is a PDB
        identifier
    :arg type: bool

    For Pfam MSA data, *label* is UniProt entry name for the protein.  You may
    also use PDB structure and chain identifiers, e.g. ``'1p38'`` or
    ``'1p38A'``, for *label* argument and UniProt entry names will be parsed
    using :func:`.parsePDBHeader` function (see also :class:`.Polymer` and
    :class:`.DBRef`).

    The order of refinements are applied in the order of arguments.  If *label*
    and *unique* is specified, sequence matching *label* will
    be kept in the refined :class:`.MSA` although it may be similar to some
    other sequence."""

    # if msa is a char array, it will be refined but label won't work
    try:
        ndim, dtype_ = msa.ndim, msa.dtype
    except AttributeError:
        try:
            arr = msa._getArray()
        except AttributeError:
            raise TypeError('msa must be a character array or an MSA instance')
        ndim, dtype_ = arr.ndim, arr.dtype
    else:
        arr, msa = msa, None

    if dtype('|S1') != dtype_:
        raise ValueError('msa must be a character array or an MSA instance')
    if ndim != 2:
        raise ValueError('msa must be a 2D array or an MSA instance')

    title = []
    cols = None

    if index is not None:
        before = arr.shape[1]
        LOGGER.timeit('_refine')
        cols = char.isalpha(arr[index]).nonzero()[0]
        arr = arr.take(cols, 1)
        title.append('index=' + str(index))
        LOGGER.report('Index refinement reduced number of columns from {0} to '
                      '{1} in %.2fs.'.format(before, arr.shape[1]), '_refine')

    if label is not None:
        if index is not None:
            LOGGER.info('An index was provided so the label will be ignored.')

        else:
            before = arr.shape[1]
            LOGGER.timeit('_refine')
            try:
                upper, lower = label.upper(), label.lower()
            except AttributeError:
                raise TypeError('label must be a string')

            if msa is None:
                raise TypeError('msa must be an MSA instance, '
                                'label cannot be used')

            index = msa.getIndex(label)
            if index is None:
                index = msa.getIndex(upper)
            if index is None:
                index = msa.getIndex(lower)

            chain = None
            if index is None and (len(label) == 4 or len(label) == 5):
                from prody import parsePDB
                try:
                    structure, header = parsePDB(label[:4], header=True)
                except Exception as err:
                    raise IOError('failed to parse header for {0} ({1})'
                                  .format(label[:4], str(err)))

                chid = label[4:].upper()
                for poly in header['polymers']:
                    if chid and poly.chid != chid:
                        continue
                    for dbref in poly.dbrefs:
                        if index is None:
                            index = msa.getIndex(dbref.idcode)
                            if index is not None:
                                LOGGER.info('{0} idcode {1} for {2}{3} '
                                            'is found in chain {4}.'.format(
                                            dbref.database, dbref.idcode,
                                            label[:4], poly.chid, str(msa)))
                                break
                        if index is None:
                            index = msa.getIndex(dbref.accession)
                            if index is not None:
                                LOGGER.info('{0} accession {1} for {2}{3} '
                                            'is found in chain {4}.'.format(
                                            dbref.database, dbref.accession,
                                            label[:4], poly.chid, str(msa)))
                                break
                if index is not None:
                    chain = structure[poly.chid]
                    resnums = chain.ca.getResnums()
    
            if index is None:
                raise ValueError('label is not in msa, or msa is not indexed')
            try:
                len(index)
            except TypeError:
                pass
            else:
                raise ValueError('label {0} maps onto multiple sequences, '
                                 'so cannot be used for refinement'.format(label))

            title.append('label=' + label)
            cols = char.isalpha(arr[index]).nonzero()[0]
            arr = arr.take(cols, 1)
            LOGGER.report('Label refinement reduced number of columns from {0} to '
                          '{1} in %.2fs.'.format(before, arr.shape[1]), '_refine')

            if chain is not None and not kwargs.get('keep', False):
                before = arr.shape[1]
                LOGGER.timeit('_refine')
                
                from prody.utilities import MATCH_SCORE, MISMATCH_SCORE, alignBioPairwise
                from prody.utilities import GAP_PENALTY, GAP_EXT_PENALTY, ALIGNMENT_METHOD

                chseq = chain.getSequence()
                algn = alignBioPairwise(pystr(arr[index].tobytes().upper()), pystr(chseq),
                                        "local",
                                        MATCH_SCORE, MISMATCH_SCORE,
                                        GAP_PENALTY, GAP_EXT_PENALTY,
                                        max_alignments=1)
                torf = []
                for s, c in zip(*algn[0][:2]):
                    if s == '-':
                        continue
                    elif c != '-':
                        torf.append(True)
                    else:
                        torf.append(False)
                torf = array(torf)
                tsum = torf.sum()
                assert tsum <= before, 'problem in mapping sequence to structure'
                if tsum < before:
                    arr = arr.take(torf.nonzero()[0], 1)
                    resnums = resnums.take(torf.nonzero()[0]-torf.nonzero()[0][0]+1)
                    LOGGER.report('Structure refinement reduced number of '
                                  'columns from {0} to {1} in %.2fs.'
                                  .format(before, arr.shape[1]), '_refine')
                else:
                    LOGGER.debug('All residues in the sequence are contained in '
                                 'PDB structure {0}.'.format(label))

                labels = msa._labels
                labels[index] = splitSeqLabel(labels[index])[0] + '/' + str(resnums[0]) + '-' + str(resnums[-1])

    from .analysis import calcMSAOccupancy, uniqueSequences

    rows = None
    if rowocc is not None:
        before = arr.shape[0]
        LOGGER.timeit('_refine')
        try:
            rowocc = float(rowocc)
        except Exception as err:
            raise TypeError('rowocc must be a float ({0})'.format(str(err)))
        assert 0. <= rowocc <= 1., 'rowocc must be between 0 and 1'

        rows = calcMSAOccupancy(arr, 'row') >= rowocc
        if index is not None:
            index = rows[:index].sum()
        rows = (rows).nonzero()[0]
        arr = arr[rows]
        title.append('rowocc>=' + str(rowocc))
        LOGGER.report('Row occupancy refinement reduced number of rows from '
                      '{0} to {1} in %.2fs.'.format(before, arr.shape[0]),
                      '_refine')

    if seqid is not None:
        before = arr.shape[0]
        LOGGER.timeit('_refine')
        unique = uniqueSequences(arr, seqid)
        if index is not None:
            unique[index] = True
        unique = unique.nonzero()[0]
        arr = arr[unique]
        title.append('seqid>=' + str(seqid))
        if rows is not None:
            rows = rows[unique]
        else:
            rows = unique
        LOGGER.report('Sequence identity refinement reduced number of rows '
                      'from {0} to {1} in %.2fs.'.format(before, arr.shape[0]),
                      '_refine')

    if colocc is not None:
        before = arr.shape[1]
        LOGGER.timeit('_refine')
        try:
            colocc = float(colocc)
        except Exception as err:
            raise TypeError('colocc must be a float ({0})'.format(str(err)))
        assert 0. <= colocc <= 1., 'colocc must be between 0 and 1'

        cols = (calcMSAOccupancy(arr, 'col') >= colocc).nonzero()[0]
        arr = arr.take(cols, 1)
        title.append('colocc>=' + str(colocc))
        LOGGER.report('Column occupancy refinement reduced number of columns '
                      'from {0} to {1} in %.2fs.'.format(before, arr.shape[1]),
                      '_refine')

    if not title:
        raise ValueError('label, index, seqid, rowocc, colocc all cannot be None')

    # depending on slicing of rows, arr may not have it's own memory
    if arr.base is not None:
        arr = arr.copy()

    if msa is None:
        return arr
    else:
        if rows is None:
            from copy import copy
            labels = copy(msa._labels)
        else:
            labels = msa._labels
            labels = [labels[i] for i in rows]
        return MSA(arr, title=msa.getTitle() + ' refined ({0})'
                   .format(', '.join(title)), labels=labels)


def mergeMSA(*msa, **kwargs):
    """Returns an :class:`.MSA` obtained from merging parts of the sequences
    of proteins present in multiple *msa* instances.  Sequences are matched
    based on protein identifiers found in the sequence labels.  Order of
    sequences in the merged MSA will follow the order of sequences in the
    first *msa* instance.  Note that protein identifiers that map to multiple
    sequences will be excluded.
    
    MSAs with different identifiers can be merged with the *ignore_ids* kwarg.
    This only works when all MSAs have the same number of sequences.
    
    :arg msa: a set of :class:`.MSA` objects to be analysed
    :type msa: list, tuple, :class:`~numpy.ndarray`
    
    :arg ignore_ids: where to ignore identifiers instead of matching them
        Default is False
    :type ignore_ids: bool
    """
    ignore_ids = kwargs.pop('ignore_ids', False)

    if len(msa) <= 1:
        raise ValueError('more than one msa instances are needed')

    try:
        arrs = [m._getArray() for m in msa]
        sets = []
        for m in msa:
            aset = set([])
            count = m.countLabel
            for label in m.iterLabels():
                if count(label) == 1:
                    aset.add(label)
            sets.append(aset)
    except AttributeError:
        raise TypeError('all msa arguments must be MSA instances')

    num_seqs = [len(aset) for aset in sets]
    sets = iter(sets)
    common = next(sets)
    if not all(array(num_seqs) == num_seqs[0]):
        LOGGER.warn("Can only ignore identifiers if all MSAs have the same number of sequences")
        return None
    
    for aset in sets:
        if not ignore_ids:
            common = common.intersection(aset)
        else:
            [common.add(item) for item in aset]
    if not common:
        LOGGER.warn("No sequences with common identifiers to merge")
        return None

    lens = [m.numResidues() for m in msa]
    rngs = [0]
    rngs.extend(cumsum(lens))
    rngs = [(start, end) for start, end in zip(rngs[:-1], rngs[1:])]

    if not ignore_ids:
        idx_arr_rng = list(zip([m.getIndex for m in msa], arrs, rngs))
        merger = zeros((len(common), sum(lens)), '|S1')
        labels = []
        for index, label in enumerate(common):
            if label not in common:
                continue
            for idx, arr, (start, end) in idx_arr_rng:
                merger[index, start:end] = arr[idx(label)]
            labels.append(label)
    else:
        idx_arr_rng = list(zip([i for i in range(len(msa))], arrs, rngs))
        merger = zeros((len(num_seqs), sum(lens)), '|S1')
        labels = msa[0].getLabels()
        for index, label in enumerate(msa[0].getLabels()):
            for i, arr, (start, end) in idx_arr_rng:
                    merger[index, start:end] = arr[index]
            labels[index] += '_' + msa[i].getLabels()[index]
            
    merger = MSA(merger, labels=labels,
                 title=' + '.join([m.getTitle() for m in msa]))
    return merger

def specMergeMSA(*msa, **kwargs):
    """Returns an :class:`.MSA` obtained from merging parts of the sequences
    of proteins present in multiple *msa* instances. Sequences are matched
    based on species section of protein identifiers found in the sequence labels.  
    Order of sequences in the merged MSA will follow the order of sequences in the
    first *msa* instance. Note that protein identifiers that map to multiple
    sequences will be excluded."""

    if len(msa) <= 1:
        raise ValueError('more than one msa instances are needed')
    lbl={}
    try:
        arrs = [m._getArray() for m in msa]
        sets = []
        labells = []
        for m in msa:
            aset = set([])
            labell = {}
            count = m.countLabel
            for label in m.iterLabels():
                lbl[label]=label.rsplit('_')[1]
                if count(label) == 1 and lbl[label] not in aset:
                    aset.add(lbl[label])
                    labell[lbl[label]]=label
            sets.append(aset)
            labells.append(labell)
    except AttributeError:
        raise TypeError('all msa arguments must be MSA instances')  
    sets = iter(sets)
    common = next(sets)
    for aset in sets:
        common = common.intersection(aset)
    if not common:
        return None
    lens = [m.numResidues() for m in msa]
    rngs = [0]
    rngs.extend(cumsum(lens))
    rngs = [(start, end) for start, end in zip(rngs[:-1], rngs[1:])]
    
    idx_arr_rng = list(zip([m.getIndex for m in msa], arrs, rngs))

    merger = zeros((len(common), sum(lens)), '|S1')
    labels = []
    for index, lbl in enumerate(common):
        for idx, arr, (start, end) in idx_arr_rng:
            merger[index, 0:start]=list(str(msa[0][msa[0].getIndex(labells[0][lbl])]))
            merger[index, start:end]=list(str(msa[1][msa[1].getIndex(labells[1][lbl])]))
        label = labells[0][lbl]

        labels.append(label)
    merger = MSA(merger, labels=labels,
                 title=' + '.join([m.getTitle() for m in msa]))
    return merger



"""This module defines classes handling individual conformations."""

from prody.measure import getRMSD
from prody.measure import Transformation

__all__ = ['Conformation', 'PDBConformation']

class Conformation(object):

    """A class to provide methods on a conformation in an ensemble.
    Instances of this class do not keep coordinate and weights data."""

    __slots__ = ['_ensemble', '_index']

    def __init__(self, ensemble, index):

        self._ensemble = ensemble
        self._index = index

    def __repr__(self):

        if self._ensemble._indices is None:
            return ('<Conformation: {0} from {1} ({2} atoms)>').format(
                    self._index, self._ensemble.getTitle(), self.numAtoms())
        else:
            return ('<Conformation: {0} from {1} (selected {2} of {3} '
                    'atoms)>').format(self._index, self._ensemble.getTitle(),
                    self.numSelected(), self.numAtoms())

    def __str__(self):

        return 'Conformation {0} from {1}'.format(
                self._index, self._ensemble.getTitle())

    def numAtoms(self):
        """Returns number of atoms."""

        return self._ensemble.numAtoms()

    def numSelected(self):
        """Returns number of selected atoms."""

        return self._ensemble.numSelected()

    def getAtoms(self):
        """Returns associated atom group."""

        return self._ensemble.getAtoms()

    def getIndex(self):
        """Returns conformation index."""

        return self._index

    def getWeights(self):
        """Returns coordinate weights for (selected) atoms."""

        ensemble = self._ensemble
        return ensemble.getWeights()

    def _getWeights(self):
        ensemble = self._ensemble
        return ensemble._getWeights()

    def getData(self, label):
        """Returns a copy of the data array associated with *label*, or **None**
        if such data is not present."""

        data = self._ensemble.getData(label)
        if data is not None:
            data = data[self._index]
        return data

    def _getData(self, label):
        """Returns data array associated with *label*, or **None** if such data
        is not present."""

        data = self._ensemble._getData(label)
        if data is not None:
            data = data[self._index]
        return data

    def getEnsemble(self):
        """Returns the ensemble that this conformation belongs to."""

        return self._ensemble

    def getCoords(self):
        """Returns a copy of the coordinates of the conformation. If a subset
        of atoms are selected in the ensemble, coordinates for selected
        atoms will be returned."""

        ensemble = self._ensemble
        if ensemble._confs is None:
            return None
        indices = ensemble._indices
        if indices is None:
            return ensemble._confs[self._index].copy()
        else:
            return ensemble._confs[self._index, indices].copy()

    def _getCoords(self):

        ensemble = self._ensemble
        if ensemble._confs is None:
            return None
        indices = ensemble._indices
        if indices is None:
            return ensemble._confs[self._index]
        else:
            return ensemble._confs[self._index, indices]

    def getDeviations(self):
        """Returns deviations from the ensemble reference coordinates.
        Deviations are calculated for (selected) atoms."""

        ensemble = self._ensemble
        if ensemble._confs is None:
            return None
        indices = ensemble._indices
        if indices is None:
            return ensemble._confs[self._index] - ensemble._coords
        else:
            return (ensemble._confs[self._index, indices].copy() -
                    ensemble._coords[indices])

    def getRMSD(self):
        """Returns RMSD from the ensemble reference coordinates. RMSD is
        calculated for (selected) atoms."""

        ensemble = self._ensemble
        indices = ensemble._indices
        if indices is None:
            return getRMSD(ensemble._coords,
                           ensemble._confs[self._index],
                           self.getWeights())[0]
        else:
            return getRMSD(ensemble._coords[indices],
                           ensemble._confs[self._index, indices],
                           self.getWeights())[0]

class PDBConformation(Conformation):

    """This class is the same as :class:`Conformation`, except that the
    conformation has a name (or identifier), e.g. PDB identifier."""

    def __repr__(self):

        if self.numSelected() == self.numAtoms():
            return ('<PDBConformation: {0} from {1} (index: {2}; '
                    '{3} atoms)>').format(
                    self._ensemble._labels[self._index],
                    self._ensemble.getTitle(), self._index, self.numAtoms())
        else:
            return ('<PDBConformation: {0} from {1} (index: {2}; '
                    'selected {3} of {4} atoms)>').format(
                        self._ensemble._labels[self._index],
                        self._ensemble.getTitle(), self._index,
                        self.numSelected(), self.numAtoms())

    def __str__(self):

        return 'PDBConformation {0} from {1}'.format(
                    self._ensemble._labels[self._index],
                    self._ensemble.getTitle())

    def getData(self, label):
        data = self._ensemble.getData(label)

        if data is None:
            return None
        return data[self._index]
    
    def setData(self, label, value):
        data = self._ensemble._getData(label)

        if data is not None:
            data[self._index] = value

    def getLabel(self):
        """Returns the label of the conformation."""

        return self._ensemble._labels[self._index]

    def setLabel(self, label):
        """Set the label of the conformation."""

        self._ensemble._labels[self._index] = str(label)

    def getSequence(self):
        """Returns the sequence of the conformation."""

        return self._ensemble.getMSA()[self._index]

    def getCoords(self):
        """Returns a copy of the coordinates of the conformation. If a subset
        of atoms are selected in the ensemble, coordinates for selected
        atoms will be returned.

        .. warning:: When there are atoms with weights equal to zero (0),
           their coordinates will be replaced with the coordinates of the
           ensemble reference coordinate set."""

        ensemble = self._ensemble
        if ensemble._confs is None:
            return None
        indices = ensemble._indices
        index = self._index
        if indices is None:
            coords = ensemble._confs[index].copy()
            which = ensemble._weights[index].flatten()==0
            coords[which] = ensemble._coords[which]
        else:
            coords = ensemble._confs[index, indices].copy()
            which = ensemble._weights[index, indices].flatten()==0
            coords[which] = ensemble._coords[indices][which]
        return coords

    _getCoords = getCoords

    def getDeviations(self):
        """Returns deviations from the ensemble reference coordinates.
        Deviations are calculated for (selected) atoms."""

        ensemble = self._ensemble
        indices = ensemble._indices
        if indices is None:
            return self.getCoords() - ensemble._coords
        else:
            return self.getCoords() - ensemble._coords[indices]

    def getRMSD(self):
        """Returns RMSD from the ensemble reference coordinates. RMSD is
        calculated for (selected) atoms."""

        ensemble = self._ensemble
        index = self._index
        indices = ensemble._indices
        if indices is None:
            return getRMSD(ensemble._coords,
                           ensemble._confs[index],
                           ensemble._weights[index])
        else:
            return getRMSD(ensemble._coords[indices],
                           ensemble._confs[index, indices],
                           ensemble._weights[index, indices])

    def getWeights(self):
        """Returns coordinate weights for (selected) atoms."""

        ensemble = self._ensemble
        indices = ensemble._indices
        if indices is None:
            return ensemble._weights[self._index].copy()
        else:
            return ensemble._weights[self._index, indices]

    def _getWeights(self):

        ensemble = self._ensemble
        indices = ensemble._indices
        if indices is None:
            return ensemble._weights[self._index]
        else:
            return ensemble._weights[self._index, indices]

    def getTransformation(self):
        """Returns the :class:`~.Transformation` used to superpose this
        conformation onto reference coordinates.  The transformation can
        be used to superpose original PDB file onto the reference PDB file."""

        if self._ensemble._trans is not None:
            return Transformation(self._ensemble._trans[self._index])




# -*- coding: utf-8 -*-
"""This module defines functions for editing normal mode data."""

import numpy as np

from prody.atomic import Atomic, AtomGroup, AtomMap, AtomSubset
from prody.atomic import Selection, SELECT, sliceAtoms, extendAtoms
from prody.measure import calcDistance
from prody.utilities import importLA, isListLike, getCoords
from prody import _PY3K, LOGGER

from .nma import NMA
from .mode import VectorBase, Mode, Vector
from .gnm import GNM
from .anm import ANM
from .pca import PCA

if not _PY3K:
    range = xrange

__all__ = ['extendModel', 'extendMode', 'extendVector',
           'sliceMode', 'sliceModel', 'sliceModelByMask', 'sliceVector',
           'reduceModel', 'reduceModelByMask', 'trimModel', 'trimModelByMask',
           'interpolateModel']


def extendModel(model, nodes, atoms, norm=False):
    """Extend a coarse grained *model* built for *nodes* to *atoms*.  *model*
    may be :class:`.ANM`, :class:`.GNM`, :class:`.PCA`, or :class:`.NMA`
    instance.  This function will take part of the normal modes for each node
    (i.e. Cα atoms) and extend it to all other atoms in the same residue.  For
    each atom in *nodes* argument *atoms* argument must contain a corresponding
    residue.  If *norm* is **True**, extended modes are normalized."""

    try:
        evecs = model._getArray()
        evals = model.getEigvals()
    except AttributeError:
        raise ValueError('model must be an NMA instance')

    if model.numAtoms() != nodes.numAtoms():
        raise ValueError('atom numbers must be the same')

    indices, atommap = extendAtoms(nodes, atoms, model.is3d())

    evecs = evecs[indices, :]
    if norm:
        evecs /= np.array([((evecs[:, i]) ** 2).sum() ** 0.5
                           for i in range(evecs.shape[1])])

    if model.is3d():
        extended = NMA('Extended ' + str(model))
    else:
        extended = GNM('Extended ' + str(model))
    extended.setEigens(evecs, evals)
    return extended, atommap


def extendMode(mode, nodes, atoms, norm=False):
    """Extend a coarse grained normal *mode* built for *nodes* to *atoms*.
    This function will take part of the normal modes for each node (i.e. Cα
    atoms) and extend it to all other atoms in the same residue.  For each atom
    in *nodes* argument *atoms* argument must contain a corresponding residue.
    Extended mode is multiplied by the square root of variance of the mode.
    If *norm* is **True**, extended mode is normalized."""

    try:
        vec = mode._getArray()
        std = mode.getVariance() ** 0.5
    except AttributeError:
        raise ValueError('mode must be a normal Mode instance')

    indices, atommap = extendAtoms(nodes, atoms, mode.is3d())
    vec = vec[indices]
    if norm:
        vec /= ((vec) ** 2).sum() ** 0.5
    else:
        vec *= std
    extended = Vector(vec, 'Extended ' + str(mode), mode.is3d())
    return extended, atommap


def extendVector(vector, nodes, atoms):
    """Extend a coarse grained *vector* for *nodes* to *atoms*.  This function
    will take part of the normal modes for each node (i.e. Cα atoms) and extend
    it to all other atoms in the same residue.  For each atom in *nodes*,
    *atoms* argument must contain a corresponding residue."""

    try:
        vec = vector._getArray()
    except AttributeError:
        raise ValueError('vector must be a Vector instance')

    if vector.numAtoms() != nodes.numAtoms():
        raise ValueError('atom numbers must be the same')

    indices, atommap = extendAtoms(nodes, atoms, vector.is3d())
    extended = Vector(vec[indices], 'Extended ' + str(vector), vector.is3d())
    return extended, atommap

def sliceVector(vector, atoms, select):
    """Returns part of the *vector* for *atoms* matching *select*.  Note that
    returned :class:`.Vector` instance is not normalized.

    :arg vector: vector instance to be sliced
    :type vector: :class:`.VectorBase`

    :arg atoms: atoms for which *vector* describes a deformation, motion, etc.
    :type atoms: :class:`.Atomic`

    :arg select: an atom selection or a selection string
    :type select: :class:`.Selection`, str

    :returns: (:class:`.Vector`, :class:`.Selection`)"""

    if not isinstance(vector, VectorBase):
        raise TypeError('vector must be a VectorBase instance, not {0}'
                        .format(type(vector)))
    if not isinstance(atoms, Atomic):
        raise TypeError('atoms must be an Atomic instance, not {0}'
                        .format(type(atoms)))
    if atoms.numAtoms() != vector.numAtoms():
        raise ValueError('number of atoms in model and atoms must be equal')

    which, sel = sliceAtoms(atoms, select)

    if vector.is3d():
        vec = Vector(vector.getArrayNx3()[which, :].flatten(),
                    '{0} slice {1}'.format(str(vector), select),
                    vector.is3d())
    else:
        vec = Vector(vector.getArray()[which].flatten(),
                    '{0} slice {1}'.format(str(vector), select),
                    vector.is3d())
    return (vec, sel)

def trimModel(model, atoms, select):
    """Returns a part of the *model* for *atoms* matching *select*. This method removes 
    columns and rows in the connectivity matrix and fix the diagonal sums. Normal modes 
    need to be calculated again after the trim.

    :arg mode: NMA model instance to be sliced
    :type mode: :class:`.NMA`

    :arg atoms: atoms for which the *model* was built
    :type atoms: :class:`.Atomic`

    :arg select: an atom selection or a selection string
    :type select: :class:`.Selection`, str

    :returns: (:class:`.NMA`, :class:`.Selection`)"""

    if not isinstance(model, NMA):
        raise TypeError('mode must be a NMA instance, not {0}'
                        .format(type(model)))
    if not isinstance(atoms, Atomic):
        raise TypeError('atoms must be an Atomic instance, not {0}'
                        .format(type(atoms)))
    if atoms.numAtoms() != model.numAtoms():
        raise ValueError('number of atoms in model and atoms must be equal')

    which, sel = sliceAtoms(atoms, select)
    nma = trimModelByMask(model, which)

    return (nma, sel)

def trimModelByMask(model, mask):
    """Returns a part of the *model* indicated by *mask*. This method removes 
    columns and rows in the connectivity matrix indicated by *mask* and fix the diagonal sums.
    Normal modes need to be calculated again after the trim.

    :arg mode: NMA model instance to be sliced
    :type mode: :class:`.NMA`

    :arg mask: an Integer array or a Boolean array where ``"True"`` indicates 
        the parts being selected 
    :type mask: list, :class:`~numpy.ndarray`

    :returns: :class:`.NMA`"""

    if not isListLike(mask):
        raise TypeError('mask must be either a list or a numpy.ndarray, not {0}'
                        .format(type(mask)))
    
    is_bool = mask.dtype is np.dtype('bool')

    if is_bool:
        if len(mask) != model.numAtoms():
            raise ValueError('number of atoms in model and mask must be equal')
        which = mask
    else:
        if mask.min() < 0 or mask.max() >= model.numAtoms():
            raise ValueError('index in mask exceeds range')
        which = np.zeros(model.numAtoms(), dtype=bool)
        which[mask] = True

    if model.is3d():
        which = np.repeat(which, 3)

    if isinstance(model, GNM):
        matrix = model._kirchhoff
    elif isinstance(model, ANM):
        matrix = model._hessian
    elif isinstance(model, PCA):
        matrix = model._cov
    
    if isinstance(model, PCA):
        ss = matrix[which, :][:, which]
        eda = PCA(model.getTitle() + ' reduced')
        eda.setCovariance(ss)
        return eda
    else:
        matrix = matrix[which, :][:, which]

        if isinstance(model, GNM):
            gnm = GNM(model.getTitle() + ' reduced')
            I = np.eye(len(matrix), dtype=bool)
            matrix[I] = - (matrix.sum(axis=0) - np.diag(matrix))
            gnm.setKirchhoff(matrix)
            return gnm
        elif isinstance(model, ANM):
            model_type = type(model)
            anm = model_type(model.getTitle() + ' reduced')
            
            n = len(matrix) // 3
            for i in range(n):
                S = np.zeros((3, 3))
                for j in range(n):
                    if i == j:
                        continue
                    S -= matrix[i*3:i*3+3, j*3:j*3+3]
                matrix[i*3:i*3+3, i*3:i*3+3] = S
            anm.setHessian(matrix)
            if hasattr(anm, 'getMembrane'):
                anm._membrane = model.getMembrane()
                anm._combined = model.getCombined()
            return anm

def sliceMode(mode, atoms, select):
    """Returns part of the *mode* for *atoms* matching *select*.  This works
    slightly different from :func:`.sliceVector`. Mode array (eigenvector) is
    multiplied by square-root of the variance along the mode.  If mode is from
    an elastic network model, variance is defined as the inverse of the
    eigenvalue.  Note that returned :class:`.Vector` instance is not
    normalized.

    :arg mode: mode instance to be sliced
    :type mode: :class:`.Mode`

    :arg atoms: atoms for which *mode* describes a deformation, motion, etc.
    :type atoms: :class:`.Atomic`

    :arg select: an atom selection or a selection string
    :type select: :class:`.Selection`, str

    :returns: (:class:`.Vector`, :class:`.Selection`)"""

    if not isinstance(mode, Mode):
        raise TypeError('mode must be a Mode instance, not {0}'
                        .format(type(mode)))
    if not isinstance(atoms, Atomic):
        raise TypeError('atoms must be an Atomic instance, not {0}'
                        .format(type(atoms)))
    if atoms.numAtoms() != mode.numAtoms():
        raise ValueError('number of atoms in model and atoms must be equal')

    which, sel = sliceAtoms(atoms, select)

    if mode.is3d():
        vec = Vector(mode.getArrayNx3()[which, :].flatten() *
                     mode.getVariance()**0.5,
                    '{0} slice {1}'.format(str(mode), select), mode.is3d())
    else:
        vec = Vector(mode.getArray()[which].flatten() *
                     mode.getVariance()**0.5,
                    '{0} slice {1}'.format(str(mode), select), mode.is3d())
    return (vec, sel)


def sliceModel(model, atoms, select, norm=False):
    """Returns a part of the *model* (modes calculated) for *atoms* matching *select*. 
    Note that normal modes are sliced instead of the connectivity matrix. Sliced normal 
    modes (eigenvectors) are not normalized unless *norm* is **True**.

    :arg mode: NMA model instance to be sliced
    :type mode: :class:`.NMA`

    :arg atoms: atoms for which the *model* was built
    :type atoms: :class:`.Atomic`

    :arg select: an atom selection or a selection string
    :type select: :class:`.Selection`, str

    :arg norm: whether to normalize eigenvectors, default **False**
    :type norm: bool

    :returns: (:class:`.NMA`, :class:`.Selection`)"""

    if not isinstance(model, NMA):
        raise TypeError('mode must be a NMA instance, not {0}'
                        .format(type(model)))
    if not isinstance(atoms, Atomic):
        raise TypeError('atoms must be an Atomic instance, not {0}'
                        .format(type(atoms)))
    if atoms.numAtoms() != model.numAtoms():
        raise ValueError('number of atoms in model and atoms must be equal')

    which, sel = sliceAtoms(atoms, select)
    nma = sliceModelByMask(model, which, norm=norm)

    return (nma, sel)

def sliceModelByMask(model, mask, norm=False):
    """Returns a part of the *model* indicated by *mask*.  Note that
    normal modes (eigenvectors) are not normalized unless *norm* is **True**.

    :arg mode: NMA model instance to be sliced
    :type mode: :class:`.NMA`

    :arg mask: an Integer array or a Boolean array where ``"True"`` indicates 
        the parts being selected 
    :type mask: list, :class:`~numpy.ndarray`

    :arg norm: whether to normalize eigenvectors, default **False**
    :type norm: bool

    :returns: :class:`.NMA`"""

    if not isListLike(mask):
        raise TypeError('mask must be either a list or a numpy.ndarray, not {0}'
                        .format(type(mask)))
    
    is_bool = mask.dtype is np.dtype('bool')

    if is_bool:
        if len(mask) != model.numAtoms():
            raise ValueError('number of atoms in model and mask must be equal')
        which = mask
    else:
        if mask.min() < 0 or mask.max() >= model.numAtoms():
            raise ValueError('index in mask exceeds range')
        which = np.zeros(model.numAtoms(), dtype=bool)
        which[mask] = True

    array = model._getArray()
    
    nma = type(model)('{0} sliced'
                    .format(model.getTitle()))
    if model.is3d():
        which = np.repeat(which, 3)

    evecs = array[which, :]
    if norm:
        evecs /= np.array([((evecs[:, i]) ** 2).sum() ** 0.5
                           for i in range(evecs.shape[1])])

    nma.setEigens(evecs, model.getEigvals())
    return nma

def reduceModel(model, atoms, select):
    """Returns reduced NMA model.  Reduces a :class:`.NMA` model to a subset of
    *atoms* matching *select*.  This function behaves differently depending on
    the type of the *model* argument.  For :class:`.ANM` and :class:`.GNM` or
    other :class:`.NMA` models, force constant matrix for system of interest
    (specified by the *select*) is derived from the force constant matrix for
    the *model* by assuming that for any given displacement of the system of
    interest, other atoms move along in such a way as to minimize the potential
    energy.  This is based on the formulation in [KH00]_.  For :class:`.PCA`
    models, this function simply takes the sub-covariance matrix for selection.

    .. [KH00] Hinsen K, Petrescu A-J, Dellerue S, Bellissent-Funel M-C, Kneller GR.
       Harmonicity in slow protein dynamics. *Chem Phys* **2000** 261:25-37.

    :arg model: dynamics model
    :type model: :class:`.ANM`, :class:`.GNM`, or :class:`.PCA`

    :arg atoms: atoms that were used to build the model
    :type atoms: :class:`.Atomic`

    :arg select: an atom selection or a selection string
    :type select: :class:`.Selection`, str

    :returns: (:class:`.NMA`, :class:`.Selection`)"""

    if not isinstance(model, NMA):
        raise TypeError('model must be an NMA instance, not {0}'
                        .format(type(model)))
    if not isinstance(atoms, (AtomGroup, AtomSubset, AtomMap)):
        raise TypeError('atoms type is not valid')
    if len(atoms) <= 1:
        raise TypeError('atoms must contain more than 1 atoms')

    if isinstance(model, GNM):
        matrix = model._kirchhoff
    elif isinstance(model, ANM):
        matrix = model._hessian
    elif isinstance(model, PCA):
        matrix = model._cov
    else:
        raise TypeError('model does not have a valid type derived from NMA')
    if matrix is None:
        raise ValueError('model matrix (Hessian/Kirchhoff/Covariance) is not '
                         'built')

    which, select = sliceAtoms(atoms, select)
    nma = reduceModelByMask(model, which)

    return nma, select

def reduceModelByMask(model, mask):
    """Returns NMA model reduced based on *mask*. 

    :arg model: dynamics model
    :type model: :class:`.ANM`, :class:`.GNM`, or :class:`.PCA`

    :arg mask: an Integer array or a Boolean array where ``"True"`` indicates 
        the parts being selected 
    :type mask: list, :class:`~numpy.ndarray`

    :returns: :class:`.NMA`"""

    if not isinstance(model, NMA):
        raise TypeError('model must be an NMA instance, not {0}'
                        .format(type(model)))
    
    if not isListLike(mask):
        raise TypeError('mask must be either a list or a numpy.ndarray, not {0}'
                        .format(type(mask)))
    
    is_bool = mask.dtype is np.dtype('bool')

    if is_bool:
        if len(mask) != model.numAtoms():
            raise ValueError('number of atoms in model and mask must be equal')
        system = mask
    else:
        if mask.min() < 0 or mask.max() >= model.numAtoms():
            raise ValueError('index in mask exceeds range')
        system = np.zeros(model.numAtoms(), dtype=bool)
        system[mask] = True

    if isinstance(model, GNM):
        matrix = model._kirchhoff
    elif isinstance(model, ANM):
        matrix = model._hessian
    elif isinstance(model, PCA):
        matrix = model._cov
    else:
        raise TypeError('model does not have a valid type derived from NMA')
    if matrix is None:
        raise ValueError('model matrix (Hessian/Kirchhoff/Covariance) is not '
                         'built')

    if model.is3d():
        system = np.repeat(system, 3)

    if isinstance(model, PCA):
        ss = matrix[system, :][:, system]
        eda = PCA(model.getTitle() + ' reduced')
        eda.setCovariance(ss)
        return eda
    else:
        matrix = _reduceModel(matrix, system)

        if isinstance(model, GNM):
            gnm = GNM(model.getTitle() + ' reduced')
            gnm.setKirchhoff(matrix)
            return gnm
        elif isinstance(model, ANM):
            anm = ANM(model.getTitle() + ' reduced')
            anm.setHessian(matrix)
            return anm

def _reduceModel(matrix, system):
    """This is the underlying function that reduces models, which shall 
    remain private. *system* is a Boolean array where **True** indicates 
    system nodes."""
    
    linalg = importLA()

    other = np.invert(system)

    ss = matrix[system, :][:, system]
    so = matrix[system, :][:, other]
    os = matrix[other, :][:, system]
    oo = matrix[other, :][:, other]

    if other.any():
        try:
            invoo = linalg.inv(oo)
        except:
            invoo = linalg.pinv(oo)
        
        matrix = ss - np.dot(so, np.dot(invoo, os))
    else:
        matrix = ss

    return matrix


def interpolateModel(model, nodes, coords, norm=False, **kwargs):
    """Interpolate a coarse grained *model* built for *nodes* to *coords*.  
    
    *model* may be :class:`.ANM`, :class:`.PCA`, or :class:`.NMA`
    instance

    :arg nodes: the coordinate set or object with :meth:`getCoords` method
        that corresponds to the model
    :type nodes: :class:`.Atomic`, :class:`~numpy.ndarray`

    :arg coords: a coordinate set or an object with :meth:`getCoords` method
        onto which the model should be interpolated
    :type coords: :class:`.Atomic`, :class:`~numpy.ndarray`
    
    This function will take the part of the normal modes for each node
    (i.e. Cα atoms) and extend it to nearby atoms. 
    
    If *norm* is **True**, extended modes are normalized.
    
    Adapted from ModeHunter as described in [JS09]_.

    .. [JS09] Stember JN, Wriggers W. Bend-twist-stretch model for coarse 
    elastic network simulation of biomolecular motion. *J Chem Phys* **2009** 131:074112.

    # Legal notice:
    # 
    # This software is copyrighted, (c) 2009-10, by Joseph N. Stember and Willy Wriggers
    # under the following terms:
    #
    # The authors hereby grant permission to use, copy, modify, and re-distribute this
    # software and its documentation for any purpose, provided that existing copyright
    # notices are retained in all copies and that this notice is included verbatim in
    # any distributions. No written agreement, license, or royalty fee is required for
    # any of the authorized uses.
    #
    # IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY FOR DIRECT,
    # INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE USE
    # OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY DERIVATIVES THEREOF, EVEN IF THE
    # AUTHORS HAVE BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
    #
    # THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES, INCLUDING,
    # BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
    # PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE IS PROVIDED ON AN "AS
    # IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE NO OBLIGATION TO PROVIDE
    # MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
    #
    # +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    """

    quiet = kwargs.pop('quiet', False)

    try:
        eigvecs = model._getArray()
        eigvals = model.getEigvals()
    except AttributeError:
        raise TypeError('model must be an NMA instance')

    if model.numAtoms() != nodes.numAtoms():
        raise ValueError('atom numbers must be the same')

    if not model.is3d():
        raise ValueError('model must be 3D')

    cg_coords = nodes.getCoords()
    len_cg = nodes.numAtoms()

    fg_coords = getCoords(coords)
    len_fg = fg_coords.shape[0]

    n_modes = model.numModes()

    eigvecs_fg = np.zeros((3*len_fg, n_modes))

    if not quiet:
        LOGGER.progress('Interpolating {0} coarse modes to fine modes...'.format(n_modes),
                        n_modes, '_prody_iterp')

    for k in np.arange(n_modes): # loop through cg eigenvecs
        coarse_grained_nms = eigvecs[:, k]
        coarse_grained_nms.shape = (-1,3)

        lmat = np.zeros((len_cg+4,len_cg+4))
        for i in np.arange(len_cg):
            for j in np.arange(len_cg):
                lmat[i,j] = np.linalg.norm(cg_coords[i]-cg_coords[j])

        for i in np.arange(len_cg):
            lmat[i,len_cg] = 1.0
            lmat[i,len_cg+1] = cg_coords[i,0]
            lmat[i,len_cg+2] = cg_coords[i,1]
            lmat[i,len_cg+3] = cg_coords[i,2]
            lmat[len_cg,i] = 1.0
            lmat[len_cg+1,i] = cg_coords[i,0]
            lmat[len_cg+2,i] = cg_coords[i,1]
            lmat[len_cg+3,i] = cg_coords[i,2]

        lmat_inv = np.linalg.inv(lmat)

        vxprime = coarse_grained_nms[:,0]
        for i in np.arange(4):
            vxprime = np.append(vxprime,0.0)

        vyprime = coarse_grained_nms[:,1]
        for i in np.arange(4):
            vyprime = np.append(vyprime,0.0)
            
        vzprime = coarse_grained_nms[:,2]
        for i in np.arange(4):
            vzprime = np.append(vzprime,0.0)

        vx = np.dot(lmat_inv,vxprime)
        vy = np.dot(lmat_inv,vyprime)
        vz = np.dot(lmat_inv,vzprime)

        nmx=np.zeros(len_fg)
        nmy=np.zeros(len_fg)
        nmz=np.zeros(len_fg)

        for j in np.arange(len_fg):
            nmx[j] += vx[len_cg] + fg_coords[j,0]*vx[len_cg+1] + fg_coords[j,1]*vx[len_cg+2] + fg_coords[j,2]*vx[len_cg+3]  
            nmy[j] += vy[len_cg] + fg_coords[j,0]*vy[len_cg+1] + fg_coords[j,1]*vy[len_cg+2] + fg_coords[j,2]*vy[len_cg+3]  
            nmz[j] += vz[len_cg] + fg_coords[j,0]*vz[len_cg+1] + fg_coords[j,1]*vz[len_cg+2] + fg_coords[j,2]*vz[len_cg+3]  

            dist_j = calcDistance(cg_coords, fg_coords[j])
            for i in np.arange(len_cg):
                nmx[j] += vx[i]*dist_j[i]
                nmy[j] += vy[i]*dist_j[i]
                nmz[j] += vz[i]*dist_j[i]

        eigvecs_fg[::3, k] = nmx
        eigvecs_fg[1::3, k] = nmy
        eigvecs_fg[2::3, k] = nmz  

        if not quiet:
            LOGGER.update(k+1, label='_prody_iterp')

    if not quiet:
        LOGGER.finish()

    if norm:
        for k in np.arange(np.size(cg_coords)):
            eigvecs_fg[k] = eigvecs_fg[k]/np.linalg.norm(eigvecs_fg[k])  
            
    interpolated = NMA('Interpolated ' + str(model))
    interpolated.setEigens(eigvecs_fg, eigvals)
    return interpolated, coords



"""Calculate conservation in an MSA using Shannon entropy."""

from ..apptools import DevelApp
from prody import LOGGER

__all__ = ['evol_conserv']

APP = DevelApp('conserv',
               help='analyze conservation using Shannon entropy')

APP.setExample(
"""This application calculates conservation using Shannon entropy for a \
refined multiple sequence alignment.  Following example will save entropy \
data and plot using default options:

    $ evol conserv piwi_refined.slx -S""", [])


APP.addArgument('msa',
    help='refined MSA file')

APP.addGroup('calc', 'calculation options')
APP.addArgument('-n', '--no-ambiguity',
    dest='ambiguity',
    help='treat amino acids characters B, Z, J, and X as non-ambiguous',
    default=True,
    action='store_false',
    group='calc')

APP.addArgument('-g', '--gaps',
    dest='omitgaps',
    help='do not omit gap characters',
    default=True,
    action='store_false',
    group='calc')

APP.addGroup('output', 'output options')
APP.addArgument('-p', '--prefix',
    dest='prefix',
    help='output filename prefix, default is '
         'msa filename with _conserv suffix',
    type=str,
    metavar='STR',
    group='output')

APP.addArgument('-f', '--number-format',
    dest='numformat', type=str, default='%12g',
    metavar='STR', help='number output format', group='output')

APP.addFigure('-S', '--save-plot',
    dest='figent',
    action='store_true',
    help='save conservation plot')


def evol_conserv(msa, **kwargs):

    import prody
    from prody import parseMSA, calcShannonEntropy, showShannonEntropy
    from prody import writeArray
    from os.path import splitext

    prefix = kwargs.get('prefix')
    if prefix is None:
        prefix, _ = splitext(msa)
        if _.lower() == '.gz':
            prefix, _ = splitext(prefix)
        prefix += '_conserv'
    msa = parseMSA(msa)
    entropy = calcShannonEntropy(msa, **kwargs)

    writeArray(prefix + '.txt',
               entropy, format=kwargs.get('numformat', '%12g'))

    if kwargs.get('figent'):
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            LOGGER.warn('Matplotlib could not be imported, '
                        'figures are not saved.')
        else:
            prody.SETTINGS['auto_show'] = False
            width = kwargs.get('figwidth', 8)
            height = kwargs.get('figheight', 6)
            figargs = kwargs.get('figargs', ())
            figure = plt.figure(figsize=(width, height))
            show = showShannonEntropy(entropy, msa=msa, *figargs)
            format = kwargs.get('figformat', 'pdf')
            figure.savefig(prefix + '.' + format, format=format,
                        dpi=kwargs.get('figdpi', 300))


APP.setFunction(evol_conserv)



# -*- coding: utf-8 -*-
"""This module defines :class:`NBExclusion` for dealing with bond information provided
by using :meth:`.AtomGroup.setNBExclusions` method."""

from numbers import Integral
import numpy as np

__all__ = ['NBExclusion']


class NBExclusion(object):

    """A pointer class for nonnonbonded exclusion exclusion atoms.  Following built-in functions are
    customized for this class:

    * :func:`iter` yields :class:`~.Atom` instances"""

    __slots__ = ['_ag', '_acsi', '_indices']

    def __init__(self, ag, indices, acsi=None):

        self._ag = ag
        self._indices = np.array(indices)
        if acsi is None:
            self._acsi = ag.getACSIndex()
        else:
            self._acsi = acsi

    def __repr__(self):

        one, two = self._indices
        names = self._ag._getNames()
        return '<NBExclusion: {0}({1})--{2}({3}) from {4}>'.format(
            names[one], one, names[two], two, str(self._ag))

    def __str__(self):

        one, two = self._indices
        names = self._ag._getNames()
        return '{0}({1})--{2}({3})'.format(
            names[one], one, names[two], two)

    def __eq__(self, other):

        return (isinstance(other, NBExclusion) and other.getAtomGroup() is self._ag
                and (np.all(other.getIndices() == self._indices) or
                     np.all(other.getIndices() == list(reversed(self._indices)))))

    def __ne__(self, other):

        return not self.__eq__(other)

    def __iter__(self):

        for index in self._indices:
            yield self._ag[index]

    def getAtomGroup(self):
        """Returns atom group."""

        return self._ag

    def getAtoms(self):
        """Returns nonbonded exclusion atoms."""

        return (self._ag[self._indices[0]], self._ag[self._indices[1]])

    def getIndices(self):
        """Returns indices of nonbonded exclusion atoms."""

        return self._indices.copy()

    def getVector(self):
        """Returns vector that originates from the first atom."""

        one, two = self._indices
        acsi = self.getACSIndex()
        return self._ag._coords[acsi, two] - self._ag._coords[acsi, one]

    def getACSIndex(self):
        """Returns index of the coordinate set."""

        acsi = self._acsi
        if acsi >= self._ag._n_csets:
            raise ValueError('{0} has fewer coordsets than assumed by {1}'
                             .format(str(self._ag), str(self)))
        return acsi

    def setACSIndex(self, index):
        """Set the coordinate set at *index* active."""

        if self._ag._coords is None:
            raise AttributeError('coordinates are not set')

        if not isinstance(index, Integral):
            raise TypeError('index must be an integer')

        n_csets = self._ag._n_csets
        if n_csets <= index or n_csets < abs(index):
            raise IndexError('coordinate set index is out of range')

        if index < 0:
            index += n_csets

        self._acsi = index


def evalNBExclusions(exclusions, n_atoms):
    """Returns an array mapping atoms to their nonbonded exclusion neighbors and an array
    that stores number of nonbonded exclusions made by each atom."""

    numexclusions = np.bincount(
        exclusions.reshape((exclusions.shape[0] * 2)))
    nbemap = np.zeros((n_atoms, numexclusions.max()), int)
    nbemap.fill(-1)
    index = np.zeros(n_atoms, int)
    for nbexclusion in exclusions:
        a, b = nbexclusion
        nbemap[a, index[a]] = b
        nbemap[b, index[b]] = a
        index[nbexclusion] += 1
    return nbemap, numexclusions


def trimNBExclusions(exclusions, indices):
    """Returns nonbonded exclusions between atoms at given indices."""

    iset = set(indices)
    exclusions = [nbexclusion for nbexclusion in exclusions if nbexclusion[0]
                    in iset and nbexclusion[1] in iset]
    if exclusions:
        newindices = np.zeros(indices.max()+1, int)
        newindices[indices] = np.arange(len(indices))
        return newindices[np.array(exclusions)]



# -*- coding: utf-8 -*-
"""This module defines functions for executing STRIDE program and parsing
its output."""

import os.path

import numpy as np

from prody.atomic import ATOMIC_FIELDS
from prody.atomic import AtomGroup
from prody.utilities import gunzip, which

from .pdbfile import parsePDB
from .localpdb import fetchPDB

__all__ = ['execSTRIDE', 'parseSTRIDE', 'performSTRIDE']


def execSTRIDE(pdb, outputname=None, outputdir=None):
    """Execute STRIDE program for given *pdb*.  *pdb* can be an identifier or
    a PDB file path.  If *pdb* is a compressed file, it will be decompressed
    using Python :mod:`gzip` library.  When no *outputname* is given, output
    name will be :file:`pdb.stride`.  :file:`.stride` extension will be
    appended automatically to *outputname*.  If :file:`outputdir` is given,
    STRIDE output and uncompressed PDB file will be written into this folder.
    Upon successful execution of :command:`stride pdb > out` command, output
    filename is returned.

    For more information on STRIDE see http://webclu.bio.wzw.tum.de/stride/.
    If you benefited from STRIDE, please consider citing [DF95]_.

    .. [DF95] Frishman D, Argos P. Knowledge-Based Protein Secondary Structure
       Assignment. *Proteins* **1995** 23:566-579."""

    stride = which('stride')
    if stride is None:
        raise EnvironmentError('command not found: stride executable is not '
                               'found in one of system paths')
    assert outputname is None or isinstance(outputname, str),\
        'outputname must be a string'
    assert outputdir is None or isinstance(outputdir, str),\
        'outputdir must be a string'
    if not os.path.isfile(pdb):
        pdb = fetchPDB(pdb, compressed=False)
    if pdb is None:
        raise ValueError('pdb is not a valid PDB identifier or filename')
    if os.path.splitext(pdb)[1] == '.gz':
        if outputdir is None:
            pdb = gunzip(pdb, os.path.splitext(pdb)[0])
        else:
            pdb = gunzip(pdb, os.path.join(outputdir,
                         os.path.split(os.path.splitext(pdb)[0])[1]))
    if outputdir is None:
        outputdir = '.'
    if outputname is None:
        out = os.path.join(outputdir,
                           os.path.splitext(os.path.split(pdb)[1])[0] +
                           '.stride')
    else:
        out = os.path.join(outputdir, outputname + '.stride')

    status = os.system('{0} {1} > {2}'.format(stride, pdb, out))
    if status == 0:
        return out


def parseSTRIDE(stride, ag):
    """Parse STRIDE output from file *stride* into :class:`~.AtomGroup`
    instance *ag*.  STRIDE output file must be in the new format used
    from July 1995 and onwards.  When *stride* file is parsed, following
    attributes are added to *ag*:

    * *stride_resnum*: STRIDE's sequential residue number, starting at the
      first residue actually in the data set.

    * *stride_phi*, *stride_psi*: peptide backbone torsion angles phi and psi

    * *stride_area*: residue solvent accessible area"""

    if not os.path.isfile(stride):
        raise IOError('{0} is not a valid file path'.format(stride))
    if not isinstance(ag, AtomGroup):
        raise TypeError('ag argument must be an AtomGroup instance')

    stride = open(stride)

    n_atoms = ag.numAtoms()
    NUMBER = np.zeros(n_atoms, int)
    AREA = np.zeros(n_atoms, float)
    PHI = np.zeros(n_atoms, float)
    PSI = np.zeros(n_atoms, float)

    ag.setSecstrs(np.zeros(n_atoms, dtype=ATOMIC_FIELDS['secondary'].dtype))
    for line in stride:
        if not line.startswith('ASG '):
            continue
        res = ag[(line[9], int(line[10:15]), line[15].strip())]
        if res is None:
            continue
        indices = res.getIndices()
        res.setSecstrs(line[24].strip())
        NUMBER[indices] = int(line[16:20])
        PHI[indices] = float(line[42:49])
        PSI[indices] = float(line[52:59])
        AREA[indices] = float(line[64:69])
    ag.setData('stride_resnum', NUMBER)
    ag.setData('stride_phi', PHI)
    ag.setData('stride_psi', PSI)
    ag.setData('stride_area', AREA)
    return ag


def performSTRIDE(pdb):
    """Perform STRIDE calculations and parse results.  STRIDE data is
    returned in an :class:`.AtomGroup` instance.  See also
    :func:`execSTRIDE` and :func:`parseSTRIDE`."""

    pdb = fetchPDB(pdb, compressed=False)
    return parseSTRIDE(execSTRIDE(pdb), parsePDB(pdb))



from os import remove
import shlex
from os.path import isfile, join, split, splitext
from prody.tests import TestCase, skipIf, skipUnless

from numpy.testing import *

from prody.utilities import importDec
dec = importDec()

from prody.tests.datafiles import TEMPDIR, pathDatafile

from prody.apps import prody_parser

from prody.tests import MATPLOTLIB, NOPRODYCMD, WINDOWS


class TestANMCommand(TestCase):

    def setUp(self):

        self.command = ('anm -e -r -o {outdir} -v -z -t all '
                        '-f %8g -d , -x .dat '
                        '-R -Q '
                        '-F png -D 120 -W 5 -H 4 ').format(outdir=TEMPDIR)
        self.suffixes = [
            '_anm_cc.png',
            '_anm.anm.npz',
            '_anm_covariance.dat',
            '_anm_cross-correlations.dat',
            '_anm_evalues.dat',
            '_anm_evectors.dat',
            '_anm_sf.png',
            '_anm_extended_all.nmd',
            '_anm.nmd',
        ]

        self.tearDown()

    @dec.slow
    @skipIf(NOPRODYCMD, 'prody command not found')
    @skipUnless(MATPLOTLIB, 'matplotlib not found')
    @skipIf(WINDOWS, 'command tests are not run on Windows')
    def testANMCommand(self):

        pdb = pathDatafile('multi_model_truncated')
        command = self.command + pdb
        prefix = splitext(split(pdb)[1])[0]

        namespace = prody_parser.parse_args(shlex.split(command))
        namespace.func(namespace)

        for suffix in self.suffixes:
            fn = join(TEMPDIR, prefix + suffix)
            self.assertTrue(isfile(fn), msg=fn+' not found')

    def tearDown(self):

        for pdb in [pathDatafile('multi_model_truncated')]:
            prefix = splitext(split(pdb)[1])[0]
            for suffix in self.suffixes:
                fn = join(TEMPDIR, prefix + suffix)
                if isfile(fn):
                    remove(fn)



"""Pfam MSA download application."""

__author__ = 'Ahmet Bakan, Anindita Dutta'

from ..apptools import DevelApp

__all__ = ['evol_fetch']

APP = DevelApp('fetch', help='fetch MSA files from Pfam')


APP.setExample(
"""Fetch MSA files from Pfam database for given Pfam ID code or
accession.

Fetch PFAM ID Cys_knot:

  $ evol fetch Cys_knot

Fetch PFAM accession with specific parameters:

  $ evol fetch PF00007 --compressed --format fasta --outname mymsa""", [])


APP.addArgument('acc',
    help='Pfam accession or ID',
    type=str)

APP.addGroup('download', 'download options')
APP.addArgument('-a', '--alignment',
    dest='alignment',
    type=str,
    default='full',
    metavar='STR',
    help='alignment type',
    choices='full seed ncbi metagenomics'.split(),
    group='download')
APP.addArgument('-f', '--format',
    dest='format',
    type=str,
    default='stockholm',
    metavar='STR',
    help='Pfam supported MSA format',
    choices='selex fasta stockholm'.split(),
    group='download')
APP.addArgument('-o', '--order',
    dest='order',
    type=str,
    default='tree',
    metavar='STR',
    help='ordering of sequences',
    choices='tree alphabetical'.split(),
    group='download')
APP.addArgument('-i', '--inserts',
    dest='inserts',
    type=str,
    default='upper',
    metavar='STR',
    help='letter case for inserts',
    choices='upper lower'.split(),
    group='download')
APP.addArgument('-g', '--gaps',
    dest='gaps',
    type=str,
    default='dashes',
    metavar='STR',
    help='gap character',
    choices='dashes dots mixed'.split(),
    group='download')
APP.addArgument('-t', '--timeout',
    dest='timeout',
    type=int,
    default=60,
    metavar='INT',
    help='timeout for blocking connection attempts',
    group='download')

APP.addGroup('output', 'output options')
APP.addArgument('-d', '--outdir',
    dest='folder',
    type=str,
    default='.',
    metavar='PATH',
    help='output directory',
    group='output')
APP.addArgument('-p', '--outname',
    dest='outname',
    type=str,
    default=None,
    metavar='STR',
    help='output filename, default is accession and alignment type',
    group='output')
APP.addArgument('-z', '--compressed',
    dest='compressed',
    action='store_true',
    help='gzip downloaded MSA file',
    group='output')

def evol_fetch(acc, **kwargs):

    import prody
    prody.fetchPfamMSA(acc, **kwargs)


APP.setFunction(evol_fetch)



"""This module contains unit tests for :mod:`~prody.proteins`."""

import os

import numpy as np
from numpy.testing import *


from prody import *
from prody import LOGGER
from prody.utilities import which
from prody.tests import TEMPDIR, unittest
from prody.tests.datafiles import *

LOGGER.verbosity = 'none'

class TestParseMMTF(unittest.TestCase):

    def setUp(self):
        """Set PDB file data and parse the PDB file."""

        self.ubi_mmtf = DATA_FILES['1ubi_mmtf']
        self.ubi_pdb = DATA_FILES['1ubi']
        self.nmr_mmtf = DATA_FILES['2k39_mmtf']
        self.biomol_pdb = DATA_FILES['3enl_pdb']
        self.biomol_mmtf = DATA_FILES['3enl_mmtf']
        self.altloc_pdb = DATA_FILES['1pwc_pdb']
        self.altloc_mmtf = DATA_FILES['1pwc_mmtf']
        

    def testUsualCase(self):
        """Test the outcome of a simple parsing scenario."""

        #simple file
        ag = parseDatafile(self.ubi_pdb['file'])
        mag = parseDatafile(self.ubi_mmtf['file'])

        self.assertIsInstance(mag, prody.AtomGroup,
            'parseMMTF failed to return an AtomGroup instance')

        self.assertEqual(mag.numAtoms(), self.ubi_mmtf['n_atoms'],
            'parseMMTF failed to parse correct number of atoms')

        self.assertEqual(mag.numCoordsets(), self.ubi_mmtf['models'],
            'parseMMTF failed to parse correct number of coordinate sets '
            '(models)')

        self.assertEqual(prody.calcRMSD(ag,mag),0,'parseMMTF has non-zero RMSD to parsePDB')
        
        # bioassembly creation
        ag = parseDatafile(self.biomol_pdb['file'], biomol=True)
        mag = parseDatafile(self.biomol_mmtf['file'], biomol=True)

        self.assertIsInstance(mag, prody.AtomGroup,
            'parseMMTF failed to return an AtomGroup instance')

        self.assertEqual(mag.numAtoms(), self.biomol_mmtf['n_atoms'],
            'parseMMTF failed to parse correct number of atoms')

        self.assertEqual(mag.numCoordsets(), self.biomol_mmtf['models'],
            'parseMMTF failed to parse correct number of coordinate sets '
            '(models)')        
        self.assertEqual(prody.calcRMSD(ag,mag),0,'parseMMTF has non-zero RMSD to parsePDB')        
        
        # altlocs
        ag = parseDatafile(self.altloc_pdb['file'],bonds=True)
        mag = parseDatafile(self.altloc_mmtf['file'],bonds=True)

        self.assertIsInstance(mag, prody.AtomGroup,
            'parseMMTF failed to return an AtomGroup instance')

        self.assertEqual(mag.numAtoms(), self.altloc_mmtf['n_atoms'],
            'parseMMTF failed to parse correct number of atoms')

        self.assertEqual(mag.numCoordsets(), self.altloc_mmtf['models'],
            'parseMMTF failed to parse correct number of coordinate sets '
            '(models)')        
        self.assertEqual(prody.calcRMSD(ag,mag),0,'parseMMTF has non-zero RMSD to parsePDB')       
        self.assertEqual(mag.numBonds(), 2, "parseMMTF has wrong number of bonds")

        ag = parseDatafile(self.altloc_pdb['file'],altloc='all',bonds=True)
        mag = parseDatafile(self.altloc_mmtf['file'],altloc='all',bonds=True)
        
        self.assertEqual(mag.numAtoms(), 3191, "parseMMTF failed to parse correct number of atoms with altloc='any'")
        # mmtf does not interleave altlocs, so can't do all atom rmsd (they don't match up)
        self.assertEqual(prody.calcRMSD(ag.name_C,mag.name_C),0,'parseMMTF has non-zero RMSD to parsePDB')       
        self.assertEqual(mag.numBonds(), 2, "parseMMTF has wrong number of bonds")
        
        ag = parseDatafile(self.altloc_pdb['file'],altloc='B',bonds=True)
        mag = parseDatafile(self.altloc_mmtf['file'],altloc='B',bonds=True)
        
        #the whole ligand is altloc A and is covalently bound
        self.assertEqual(mag.numAtoms(), 3109, "parseMMTF failed to parse correct number of atoms with altloc='any'")
        self.assertEqual(prody.calcRMSD(ag.name_C,mag.name_C),0,'parseMMTF has non-zero RMSD to parsePDB')       
        self.assertEqual(mag.numBonds(), 1, "parseMMTF has wrong number of bonds with altloc='B'")        
        
        # multi-model file (currently fails)
        mag = parseDatafile(self.nmr_mmtf['file'])

        self.assertIsInstance(mag, prody.AtomGroup,
            'parseMMTF failed to return an AtomGroup instance')

        self.assertEqual(mag.numAtoms(), self.nmr_mmtf['n_atoms'],
            'parseMMTF failed to parse correct number of atoms')

        self.assertEqual(mag.numCoordsets(), self.nmr_mmtf['models'],
            'parseMMTF failed to parse correct number of coordinate sets '
            '(models)')        

 
